"""Miscellaneous functions that are too small to have own subpackage like for example formatting with black.
"""

from mypythontools_cicd.misc.misc_internal import reformat_with_black

__all__ = ["reformat_with_black"]
